from .BIP39_List import BIP39_List
from hashlib import sha256
from typing import List

wordlist = BIP39_List()

class Mnemonic:
    """
    Mnemonic class for holding a BIP44 seed phrase. The 24 words represent 264
    bits. The first 256 bits are 24 words from a BIP39 word list. The final 8
    bits represent a checksum, which is the first 8 bits of the sha256 hash of
    the first 260 bits of the mnemonic.
    """
    def __init__(self, language: str) -> None:
        """
        Initializes a new instance of the Mnemonic class, and sets the language
        attribute. Raises a ValueError if the language is not in the current
        list of languages.
        """
        if language in BIP39_List.LANGUAGE_LIST:
            self.language = language
            self.seed = bytes(32)
            self.checksum = bytes(1)
        else:
            raise ValueError(f"{language} not in the current language list.")


    def set_word(self, index: int, word: str) -> None:
        """
        Sets a word in this Mnemonic class instance according to the zero-based
        array and word given. Raises errors if the index is out of range or the
        language is not in the current list of languages.
        """
        if not isinstance(index, int):
            raise TypeError("The index argument given is not an int.")

        if index < 0 or index > 23:
            raise IndexError("The index argument given is out of bounds.")

        if self.language not in wordlist.get_language(word):
            raise ValueError(f"{word} not found in {self.language} list.")

        # Convert the seed into binary text.
        seed_num = int.from_bytes(self.seed, "big")
        seed_bin_text = format(seed_num, "0264b")

        # Switch out the bits for the given word argument.
        word_index = int(word_bin_text, base=2)
        word_bin_text = format(word_index, "011b")
        char_count_left = index * 11 - 1
        char_right_position = (index + 1) * 11

        bin_text_left = seed_bin_text[:char_count_left]
        bin_text_right = seed_bin_text[char_right_position:]
        new_seed_bin_text = bin_text_left + word_bin_text + bin_text_right
        new_seed_bin = int(new_seed_bin_text, base=2).to_bytes(33, "big")

        # Recalculate the checksum if the last word, which contains the
        # checksum, is not the word being set.
        if index != 23:
            new_seed_word_section = new_seed_bin[:32]
            checksum = sha256(new_seed_word_section)[:1]
            new_seed_bin = new_seed_word_section + checksum


    def is_valid_phrase(phrase: List[str], language: str) -> bool:
        """
        Returns true if the given mnemonic phrase has words that exist in the
        given language's word list, and the checksum is valid. Raises error if
        the language given is not in the current language list.
        """
        if language not in BIP39_List.LANGUAGE_LIST:
            raise ValueError(f"{language} not in the current language list.")

        is_valid_wordcount = len(phrase) == 24
        has_valid_words = all(
            [language in wordlist.get_language(word) for word in phrase]
            )

        if not is_valid_wordcount or not has_valid_words:
            return False

        indices = [wordlist.get_word_index(word, language) for word in phrase]
        bin_text_chunks = [format(index, "011b") for index in indices]
        bin_text = "".join(bin_text_chunks)

        # The seed is the first 256 bits. The checksum is the last 8 bits.
        seed_val = int(bin_text[:256], base=2).to_bytes(32, "big")
        checksum = int(bin_text[256:264], base=2).to_bytes(1, "big")

        # The checksum is the first byte of the sha256 hash of the prior 32
        # bytes.
        recalculated_checksum = sha256(seed_val).digest()[:1]
        is_valid_checksum = checksum == recalculated_checksum

        return is_valid_checksum



import os
import lagrange
from typing import List
from .point import Point
from .polynomial import Polynomial

# The field size is a prime number that should be near the max value of the
# secret key. The BIP39 24-word seed phrase creates the largest seed phrase,
# which is 256 bits, plus an 8-bit checksum.
PRIME_MODULUS = 2 ** 256 - 2 ** 32 - 977


def create_shares(threshold: int, sharecount: int, key: bytes) -> List[Point]:
    """
    Splits a secret key into a (k, n) threshold scheme according to the Shamir
    Secret Sharing (SSS) system. The secret key can be recovered with any
    combination of k number of shares, but no information is revealed about the
    secret key, even with k - 1 shares.
    """
    # Create a polynomial of k - 1 degrees and set the x^0 coefficient to the
    # secret key.
    key_num = int.from_bytes(key, "big")
    polynomial = Polynomial(key_num)

    # Add random 240-bit coefficients.
    for i in range(threshold - 1):
        random_coefficient = int.from_bytes(os.urandom(32), "big")
        polynomial.coefficients.append(random_coefficient)

    # Create shares based on x = 1, x = 2, ... x = (k - 1)
    shares = []

    for i in range(sharecount):
        point = Point()
        point.X = i + 1
        point.Y = polynomial.solve(point.X, PRIME_MODULUS)
        shares.append(point)

    return shares


def recover_key(shares: List[Point]) -> bytes:
    """
    Takes a list of Point objects, and uses Lagrange interpolation to find the
    Y-intercept, which is the secret key. If incorrect or too few Point objects
    are provided, an incorrect result will be returned. There is no error
    checking.
    """
    point_list = [(share.X, share.Y) for share in shares]
    key_int = lagrange.interpolate(point_list, PRIME_MODULUS)
    key_bin = key_int.to_bytes(32, "big")

    return key_bin



from .bitcoinshamir import *
from .mnemonic import Mnemonic
from .share import Share
from .point import Point
from .polynomial import Polynomial
from .BIP39_List import BIP39_List

wordlist = BIP39_List()


import os
from typing import List

class BIP39_List:
    """
    BIP39_List class for loading BIP39 word lists.
    """
    LANGUAGE_LIST = [
        "chinese_simplified",
        "chinese_traditional",
        "czech",
        "english",
        "french",
        "italian",
        "japanese",
        "korean",
        "portuguese",
        "spanish"
    ]

    def __init__(self) -> None:
        """
        Creates a new BIP39_List class, and loads the word lists from the
        WordLists directory.
        """
        folder = os.path.join(os.path.dirname(__file__), "WordLists")

        chinese_simp_path = os.path.join(folder, "chinese_simplified.txt")
        chinese_trad_path = os.path.join(folder, "chinese_traditional.txt")
        czech_path = os.path.join(folder, "czech.txt")
        english_path = os.path.join(folder, "english.txt")
        french_path = os.path.join(folder, "french.txt")
        italian_path = os.path.join(folder, "italian.txt")
        japanese_path = os.path.join(folder, "japanese.txt")
        korean_path = os.path.join(folder, "korean.txt")
        portuguese_path = os.path.join(folder, "portuguese.txt")
        spanish_path = os.path.join(folder, "spanish.txt")

        with open(chinese_simp_path, "r") as f:
            self.ChineseSimplified = [line.strip() for line in f.readlines()]

        with open(chinese_trad_path, "r") as f:
            self.ChineseTraditional = [line.strip() for line in f.readlines()]

        with open(czech_path, "r") as f:
            self.Czech = [line.strip() for line in f.readlines()]

        with open(english_path, "r") as f:
            self.English = [line.strip() for line in f.readlines()]

        with open(french_path, "r") as f:
            self.French = [line.strip() for line in f.readlines()]

        with open(italian_path, "r") as f:
            self.Italian = [line.strip() for line in f.readlines()]

        with open(japanese_path, "r") as f:
            self.Japanese = [line.strip() for line in f.readlines()]

        with open(korean_path, "r") as f:
            self.Korean = [line.strip() for line in f.readlines()]

        with open(portuguese_path, "r") as f:
            self.Portuguese = [line.strip() for line in f.readlines()]

        with open(spanish_path, "r") as f:
            self.Spanish = [line.strip() for line in f.readlines()]


    def get_language(self, word: str) -> List[str]:
        """
        Searches each word list for the given word and returns the string name
        of each language that contains the given word. If the word is not
        contained in any of the word lists, an empty list is returned.
        """
        languages = []

        if word in self.ChineseSimplified:
            languages.append("chinese_simplified")

        if word in self.ChineseTraditional:
            languages.append("chinese_traditional")

        if word in self.Czech:
            languages.append("czech")

        if word in self.English:
            languages.append("english")

        if word in self.French:
            languages.append("french")

        if word in self.Italian:
            languages.append("italian")

        if word in self.Japanese:
            languages.append("japanese")

        if word in self.Korean:
            languages.append("korean")

        if word in self.Portuguese:
            languages.append("portuguese")

        if word in self.Spanish:
            languages.append("spanish")

        return languages


    def get_word_list(self, language: str) -> List[str]:
        """
        Returns the word list based on the given language. Raises an error if
        the given language is not in the current language list.
        """
        if language == "chinese_simplified":
            return self.ChineseSimplified
        elif language == "chinese_traditional":
            return self.ChineseTraditional
        elif language == "czech":
            return self.Czech
        elif language == "english":
            return self.English
        elif language == "french":
            return self.French
        elif language == "italian":
            return self.Italian
        elif language == "japanese":
            return self.Japanese
        elif language == "korean":
            return self.Korean
        elif language == "portuguese":
            return self.Portuguese
        elif language == "spanish":
            return self.Spanish
        else:
            raise ValueError(f"{language} is not in the current language list.")


    def get_word(self, index: str, language: int) -> str:
        """
        Gets the word based on the given language and zero-based index of the
        word. Raises an error if the language is not in the current language
        list or if the index is out of bounds.
        """
        if language not in self.LANGUAGE_LIST:
            raise ValueError(f"{language} is not in the current language list.")

        if not isinstance(index, int):
            raise TypeError("The given index argument was not of type int.")

        if index > 2047 or index < 0:
            raise IndexError(f"The index, {index}, is out of range.")

        return self.get_word_list(language)[index]


    def get_word_index(self, word: str, language: str) -> int:
        """
        Gets the index of the given word from the given language. Raises an
        error if the language is not in the current language list or if the word
        does not exist within the given language's word list.
        """
        if language not in self.LANGUAGE_LIST:
            raise ValueError(f"{language} is not in the current language list.")

        if language not in self.get_language(word):
            raise ValueError(f"{word} not in the {language} word list.")

        return self.get_word_list(language).index(word)
        


from .BIP39_List import BIP39_List
from typing import List
from hashlib import sha256

wordlist = BIP39_List()

class Share:
    """
    Share class for holding a single share of a Shamir (k of n) threshold
    scheme. Each share is a coordinate, with an (x, y) value, where x is an
    integer, greater than 0. The secret key is assumed to be the y value at
    x = 0. One share has 264 bits total, represented as 24 words. Each word is
    11 bits. The first 240 bits encode the Y value of the share. The next 8 bits
    are the original seed's checksum. The next 3 bits are the threshold starting
    at 2, xor the first 3 bits of the share's final checksum. The next 5 bits
    are the X value starting at 1, xor the last 5 bits of the share's final
    checksum. The final 8 bits is the share's checksum of all 256 bits preceding
    it, which is the first 8 bits of the sha256 hash.

    [240: Y value]
    [8: Seed checksum]
    [3: Threshold xor checksum]
    [5: X value xor checksum]
    [8: Checksum]
    """
    def __init__(
            self, seed_checksum: bytes,
            X: int, Y: int, threshold: int) -> None:
        """
        Initializes an instance of the Share class, and assignes values for the
        (x, y) values, as well as the threshold and checksums. The threshold
        should be a value between the inclusive bounds of 2 and 9. The X value
        should be a value between the inclusive bounds of 1 and 32.
        """
        if not isinstance(threshold, int):
            raise TypeError("The given threshold argument is not of type int.")

        if threshold < 2 or threshold > 9:
            raise ValueError("The given index argument is out of bounds.")

        if not isinstance(X, int):
            raise TypeError("The given X argument is not of type int.")

        if X < 1 or X > 32:
            raise ValueError("The given X argument is out of bounds.")
        
        if len(seed_checksum) < 1:
            raise ValueError("The given seed_checksum was not at least 1 byte.")

        self.seed_checksum = seed_checksum[:1]
        self.X = X
        self.Y = Y
        self.threshold = threshold
    

    @classmethod
    def from_mnemonic_phrase(cls, phrase: List[str], language: str) -> "Share":
        """
        Returns an instance of a Share class according to the given mnemonic
        phrase. Raises an error if the language is not in the current language
        list, the mnemonic phrase has invalid words, or has the wrong number of
        words.
        """
        if language not in BIP39_List.LANGUAGE_LIST:
            raise ValueError(f"{language} not in the current language list.")

        if not isinstance(phrase, list):
            raise TypeError("The given phrase was not of the list[str] type.")
        
        if len(phrase) != 24:
            raise ValueError("The given phrase did not have 24 words.")

        for word in phrase:
            if language not in wordlist.get_language(word):
                raise ValueError(f"\"{word}\" not in {language} word list.")

        indices = [wordlist.get_word_index(word, language) for word in phrase]
        bin_text_chunks = [format(index, "011b") for index in indices]
        bin_text = "".join(bin_text_chunks)

        # The Y value is the first 240 bits.
        y_val = int(bin_text[:240], base=2)

        # The original seedphrase checksum is the next 8 bits.
        seed_checksum = int(bin_text[240:248], base=2).to_bytes(1, "big")

        # The threshold and X value are the next 8 bits. These values still have
        # an xor of the final checksum applied, which will need to be removed.
        threshold_x_xor = int(bin_text[248:256]).to_bytes(1, "big")

        # The final checksum is the last 8 bits.
        hash_byte = int(bin_text[256:264], base=2).to_bytes(1, "big")

        # Remove xor from threshold and X value
        threshold_x = threshold_x_xor ^ hash_byte
        threshold_x_bin_text = format(int.from_bytes(threshold_x, "big"), "08b")

        # The first 3 bits are the threshold. The threshold starts with a value
        # of 2, encoded as 0, so the threshold is 2 plus the encoded value.
        threshold = int(threshold_x_bin_text[:3], base=2) + 2
        
        # The last 5 bits are the X value. The X value starts with a value of 1,
        # encoded as 0, so the X value is 1 plus the encoded value.
        x_val = int(threshold_x_bin_text[3:8], base=2) + 1

        return cls(seed_checksum, x_val, y_val, threshold)


    def to_bytes(self) -> bytes:
        """
        returns the 33-byte representation of the current Share instance.
        """
        y_bin = self.Y.to_bytes(30, "big")

        # The X value starts at 1, which is encoded as 0.
        x_bin = (self.X - 1).to_bytes(1, "big")

        # The threshold starts at 2, which is encoded as 0.
        threshold_bin = (self.threshold - 2).to_bytes(1, "big")

        threshold_x_string = format(threshold_bin, "03b") + format(x_bin, "05b")
        threshold_x_bin = int(threshold_x_string, base=2).to_bytes(1, "big")

        initial_bin = y_bin + self.seed_checksum + threshold_x_bin
        hash_byte = sha256(initial_bin).digest()[:1]

        threshold_x_xor = threshold_x_bin ^ hash_byte

        return y_bin + self.seed_checksum + threshold_x_xor + hash_byte


    def is_valid(self) -> bool:
        pass